import logging
import os
import subprocess
from functools import update_wrapper
from shutil import rmtree

import click
from click.exceptions import Exit

from webdomains import __version__
from webdomains.core import ConfigurationError, WebDomain, reload_nginx
from webdomains.utils import LogPipe, setup_click_logger

logger = setup_click_logger(logging.getLogger(__name__))
logger.setLevel(logging.DEBUG)
logger.propagate = False


# HELPERS
# -----------------------------------------------------------------------------


def die(message: str, code: int = 1):
    """Print an error message and exit."""
    logger.critical(message)
    raise Exit(code)


def mkdir(path: str, **kwargs):
    """Create the directory `path` or change its mode if it exists."""
    if not os.path.isdir(path):
        logger.info("Creating directory '%s'…", path)
        os.mkdir(path, **kwargs)
    elif 'mode' in kwargs:
        logger.info("Changing mode of '%s'…", path)
        os.chmod(path, kwargs['mode'])


def reload_nginx_or_die(callback=None):
    """
    Try to reload the NGINX server and die if an error occurs - after calling
    the `callback` method.
    """
    try:
        logger.info("Reloading NGINX server…")
        reload_nginx()
    except subprocess.SubprocessError as e:
        if callable(callback):
            callback()
        die(str(e))


# COMMANDS
# -----------------------------------------------------------------------------


def domain_argument(f):
    """Add a `domain` argument which will be passed as `WebDomain` object."""

    @click.argument('domain', metavar='DOMAIN_NAME')
    def new_func(domain, *args, **kwargs):
        if not isinstance(domain, WebDomain):
            domain = WebDomain(domain)
        return f(domain, *args, **kwargs)

    return update_wrapper(new_func, f)


@click.group()
@click.option(
    '-q', '--quiet', is_flag=True, help="Suppress non-error messages."
)
@click.version_option(version=__version__, message='%(version)s')
def cli(quiet: bool = False):
    """Manage Web domains served by this host."""
    if quiet:
        logger.setLevel(logging.ERROR)


@cli.command(name='list')
@click.option(
    '-e/-d',
    '--enabled/--disabled',
    default=None,
    help="List enabled or disabled domains only.",
)
def list_webdomains(enabled: bool = None):
    """List existing domains."""
    if enabled is True:

        def _handle_domain(domain):
            if domain.is_enabled:
                click.echo(domain)

    elif enabled is False:

        def _handle_domain(domain):
            if not domain.is_enabled:
                click.echo(domain)

    else:

        def _handle_domain(domain):
            if domain.is_enabled:
                click.echo(domain)
            else:
                click.echo("{} (disabled)".format(domain))

    with os.scandir(WebDomain.nginx_sites_available_dir) as it:
        for entry in it:
            if entry.is_file() and not (
                entry.name.startswith('.') or entry.name == 'default'
            ):
                _handle_domain(WebDomain(entry.name))


@cli.command(name='info', short_help="Show domain informations.")
@domain_argument
def info_webdomain(domain: WebDomain):
    """Show informations about the domain DOMAIN_NAME."""
    if not domain.exists:
        die("Domain does not exist.")

    try:
        domain.refresh_from_conf()
    except ConfigurationError as e:
        logger.warning(str(e))
        die("Unable to load the domain's NGINX configuration.")

    click.echo("enabled: {}".format("yes" if domain.is_enabled else "no"))
    click.echo("server_name: {}".format(domain.server_name))
    click.echo("protocol: {}".format(domain.protocol.value))

    if domain.has_https:
        click.echo("certs: {}".format("yes" if domain.has_certs else "no"))

    click.echo("template_name: {}".format(domain.template_name))


@cli.command(name='add', short_help="Add a domain.")
@click.option(
    '-n',
    '--alternative-name',
    multiple=True,
    metavar="DOMAIN_NAME",
    help="Alternative server name (can be provided multiple times).",
)
@click.option(
    '-h', '--http-only', is_flag=True, help="Do not serve over HTTPS."
)
@click.option(
    '-s', '--https-only', is_flag=True, help="Redirect HTTP to HTTPS."
)
@click.option(
    '-e', '--enable', is_flag=True, help="Enable the created domain."
)
@click.option(
    '-f', '--force', is_flag=True, help="Overwrite existing configuration."
)
@click.option(
    '-t', '--template-name', metavar="NAME", help="The template name to use."
)
@domain_argument
@click.pass_context
def add_webdomain(
    ctx: click.Context,
    domain: WebDomain,
    alternative_name: tuple = None,
    http_only: bool = False,
    https_only: bool = False,
    enable: bool = False,
    force: bool = False,
    template_name: str = None,
):
    """
    Create the NGINX configuration and directories for the domain DOMAIN_NAME.

    DOMAIN_NAME will be the principal server name used for all other commands.
    It is possible to provide alternative server names for the domain with
    `--alternative-name`.

    By default, the domain will be served over HTTP and HTTPS. It is possible
    to restrict the protocol with `--http-only` or `--https-only`. Note that if
    the SSL/TLS certificates of the domain does not exist yet, the domain will
    be served over HTTP only at first. They will be automatically generated
    using dehydrated when the domain is enabled.
    """
    if alternative_name:
        domain.alternative_names = tuple(alternative_name)

    if http_only and https_only:
        raise click.UsageError(
            "--http-only and --https-only are not compatible.", ctx
        )
    elif http_only:
        domain.protocol = WebDomain.Protocol.HTTP
    elif https_only:
        domain.protocol = WebDomain.Protocol.HTTPS

    if domain.exists:
        if not force:
            die(
                "Domain already exists. "
                "Use --force if you want to regenerate its configuration."
            )
        logger.warning("NGINX configuration will be overwritten.")

    try:
        logger.info("Generating NGINX configuration…")
        domain.generate_conf(template_name)
    except OSError as e:
        die(str(e))

    mkdir(domain.local_conf_dir, mode=0o755)
    mkdir(domain.log_dir, mode=0o750)

    if enable:
        ctx.invoke(enable_webdomain, domain=domain)


@cli.command(name='remove', short_help="Remove a domain.")
@click.option(
    '-c', '--clean', is_flag=True, help="Delete all associated directories."
)
@click.option(
    '-f', '--force', is_flag=True, help="Remove even if the domain is enabled."
)
@domain_argument
@click.pass_context
def remove_webdomain(
    ctx: click.Context,
    domain: WebDomain,
    clean: bool = False,
    force: bool = False,
):
    """
    Remove the NGINX configuration of the domain DOMAIN_NAME.

    If `--clean` is given, the local NGINX configuration and logs of the domain
    will also be deleted.
    """
    if domain.is_enabled:
        if not force:
            die(
                "Domain is currently enabled. "
                "Use --force if you want to remove it nevertheless."
            )
        ctx.invoke(disable_webdomain, domain=domain)

    if domain.exists:
        logger.info("Deleting NGINX configuration…")
        domain.conf_path.unlink()
    else:
        logger.debug("The NGINX configuration does not exist.")

    if clean:
        for path in (domain.local_conf_dir, domain.log_dir):
            if os.path.isdir(path):
                logger.info("Deleting directory « %s »…", path)
                rmtree(path)


@cli.command(name='enable', short_help="Enable a domain.")
@domain_argument
def enable_webdomain(domain: WebDomain):
    """
    Enable the NGINX configuration of the domain DOMAIN_NAME and generate as
    needed its SSL/TLS certificates.
    """
    if not domain.exists:
        die("Domain does not exist.")

    try:
        domain.refresh_from_conf()
    except ConfigurationError as e:
        logger.warning(str(e))
        logger.die("Domain's NGINX configuration could not be loaded.")

    need_certs_generation = domain.has_https and not domain.has_certs

    if domain.is_enabled:
        logger.debug("Domain's NGINX configuration is already enabled.")

        if not need_certs_generation:
            return
    else:
        logger.info("Enabling domain's NGINX configuration…")
        domain.enable()

        reload_nginx_or_die(domain.disable)

    if need_certs_generation:
        try:
            logger.info("Generating SSL/TLS certificates for the domain…")
            with LogPipe(logging.DEBUG, logger=logger) as stdout:
                domain.generate_certs(stdout=stdout)
        except subprocess.CalledProcessError as e:
            die(
                "Unable to generate SSL/TLS certificates:\n{}".format(
                    e.stderr.decode().strip()
                )
            )

        logger.info("Regenerating NGINX configuration…")
        domain.generate_conf()

        reload_nginx_or_die(domain.disable)


@cli.command(name='disable', short_help="Disable a domain.")
@domain_argument
def disable_webdomain(domain: WebDomain):
    """Disable the NGINX configuration of the domain DOMAIN_NAME."""
    if not domain.exists:
        die("Domain does not exist.")

    if not domain.is_enabled:
        logger.debug("Domain's NGINX configuration is already disabled.")
        return

    logger.info("Disabling domain's NGINX configuration…")
    domain.disable()

    try:
        logger.info("Reloading NGINX server…")
        reload_nginx()
    except subprocess.SubprocessError as e:
        logger.warning(str(e))


if __name__ == '__main__':  # pragma: no cover
    cli()



import os
import re
import subprocess
from enum import Enum
from pathlib import Path

from jinja2 import Environment, FileSystemLoader, TemplateNotFound

from .utils import cached_property


class ConfigurationError(Exception):
    """An error in the configuration, optionally located at `conf_path`."""

    def __init__(self, message: str, conf_path: Path = None):
        self.message = message
        self.conf_path = conf_path


# NGINX HELPERS
# -----------------------------------------------------------------------------


def get_nginx_directive_value(name: str, conf: str) -> str:
    """Retrieve a directive value from the given NGINX configuration."""
    match = re.search(r'^[ \t]*%s[ \t]+([^;]+);' % name, conf, re.MULTILINE)
    if match is None:
        raise ConfigurationError("`%s` directive is not defined." % name)
    return match.group(1)


def reload_nginx():
    """Reload NGINX server if the configuration is valid."""
    try:
        subprocess.run(
            ['nginx', '-t', '-q'], stderr=subprocess.PIPE, check=True
        )
    except subprocess.CalledProcessError as e:
        raise subprocess.SubprocessError(
            "The NGINX configuration is invalid:\n{}".format(
                e.stderr.decode().strip()
            )
        )
    else:
        proc = subprocess.run(['systemctl', '-q', 'reload', 'nginx.service'])
        if proc.returncode:
            raise subprocess.SubprocessError(
                "The unit 'nginx.service' fails to reload."
            )


# BASE CLASS AND MIXINS
# -----------------------------------------------------------------------------


class BaseDomain:
    """
    Base class for all domain types.

    A domain is defined by a server name. It can be enabled or disabled, and
    its main configuration can be generated or retrieved.

    Subclasses must define at least the `conf_path` and `enabled_conf_path`
    properties.
    """

    #: Ordered list of directories in which templates are looked for.
    templates_dir = [
        '/etc/webdomains/templates',
        os.path.join(os.path.dirname(__file__), 'templates'),
    ]

    #: The template name to use for the configuration.
    template_name = 'server.conf'

    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return self.name

    # DYNAMIC PROPERTIES

    @property
    def conf_path(self) -> Path:
        """The path to the main configuration of the domain."""
        raise NotImplementedError(
            "Subclasses of Domain must provide 'conf_path' property."
        )

    @property
    def enabled_conf_path(self) -> Path:
        """The path to the enabled main configuration of the domain."""
        raise NotImplementedError(
            "Subclasses of Domain must provide 'enabled_conf_path' property."
        )

    @property
    def exists(self) -> bool:
        """Whether the domain is defined."""
        return self.conf_path.exists()

    @property
    def is_enabled(self) -> bool:
        """Whether the domain is enabled."""
        return self.enabled_conf_path.exists()

    # PUBLIC METHODS

    def enable(self):
        """Enable the main configuration of the domain."""
        self.enabled_conf_path.symlink_to(self.conf_path)

    def disable(self):
        """Disable the main configuration of the domain."""
        self.enabled_conf_path.unlink()

    def generate_conf(self, template_name: str = None, **kwargs):
        """
        Overwrite the main configuration with the rendered configuration
        template for the domain.
        """
        if template_name:
            self.template_name = template_name
        try:
            template = self.get_template_environment().get_template(
                self.template_name
            )
        except TemplateNotFound:
            raise OSError(
                "Unable to find the template '%s' in %s."
                % (self.template_name, repr(self.templates_dir))
            )
        context = self.get_conf_context_data(**kwargs)
        self.conf_path.write_text(template.render(**context))

    def refresh_from_conf(self):
        """Retrieve properties from the main configuration of the domain."""
        self.update_properties_from_conf(self.conf_path.read_text())

    # OVERRIDABLE METHODS

    def get_template_environment(self, **kwargs) -> Environment:
        """Return the Jinja environment to use for template rendering."""
        options = {
            'loader': FileSystemLoader(self.templates_dir),
            'trim_blocks': True,
            'keep_trailing_newline': True,
        }
        options.update(kwargs)
        return Environment(**options)

    def get_conf_context_data(self, **kwargs) -> dict:
        """Return the template context for the domain's configuration."""
        kwargs.setdefault('domain', self)
        return kwargs

    def update_properties_from_conf(self, conf: str):
        """
        Retrieve and update the domain properties with values found in the
        given configuration.
        """
        self.template_name = self.get_property_from_conf(
            conf, 'template_name', r'.+'
        )

    def get_property_from_conf(
        self, conf: str, name: str, pattern=r'.*'
    ) -> str:
        """Retrieve a property's value from the given configuration."""
        match = re.search(
            r'^## %s=(%s)$' % (name, pattern), conf, re.MULTILINE
        )
        if match is None:
            raise ConfigurationError(
                "`%s` variable is not defined." % name, self.conf_path
            )
        return match.group(1)


class AlternativeServerNamesMixin:
    """
    Provide a way to define alternative server names of a domain.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._alternative_names = tuple()

    # PROPERTIES

    @property
    def alternative_names(self) -> tuple:
        """Return the alternative server names of the domain."""
        return self._alternative_names

    @alternative_names.setter
    def alternative_names(self, value: tuple):
        if not isinstance(value, tuple):
            raise ValueError(
                "Alternative server names must be given as a tuple."
            )
        self._alternative_names = value

    # DYNAMIC PROPERTIES

    @property
    def server_name(self) -> str:
        """The server name(s) of the domain."""
        return ' '.join([self.name] + list(self.alternative_names))


class DehydratedDomainMixin(AlternativeServerNamesMixin):
    """
    Provide a way to manage the domain's SSL/TLS certificates using dehydrated.
    """

    dehydrated_certs_dir = Path('/var/lib/dehydrated/certs')
    dehydrated_domains_path = Path('/etc/dehydrated/domains.txt')

    # DYNAMIC PROPERTIES

    @cached_property
    def certs_dir(self) -> Path:
        """The directory path to SSL/TLS certificates of the domain."""
        return self.dehydrated_certs_dir / self.name

    @property
    def has_certs(self) -> bool:
        """Whether SSL/TLS certificates of the domain exist."""
        return (self.certs_dir / 'privkey.pem').exists()

    # PUBLIC METHODS

    def generate_certs(self, stdout=None, stderr=subprocess.PIPE):
        """Generate the domain's SSL/TLS certificates using dehydrated."""
        subprocess.run(
            ['dehydrated', '--cron', '--domain', self.server_name],
            check=True,
            stdout=stdout,
            stderr=stderr,
        )

        self._update_dehydrated_domains()

    def enable(self):
        """
        Enable the main configuration of the domain and ensure that its server
        names are in the dehydrated domains' list.
        """
        super().enable()

        if self.has_https and self.has_certs:
            self._update_dehydrated_domains()

    def disable(self):
        """
        Disable the main configuration of the domain and ensure that its server
        names are not in the dehydrated domains' list.
        """
        super().disable()

        if self.has_certs:
            self._update_dehydrated_domains(remove=True)

    # INTERNAL METHODS

    def _update_dehydrated_domains(self, remove=False):
        """Update the dehydrated domains' list for the domain."""
        if not self.dehydrated_domains_path.is_file():
            if not remove:
                with open(self.dehydrated_domains_path, 'x') as f:
                    f.write('%s\n' % self.server_name)
            return

        domain_name_re = re.compile(r'(?:^|.+ ){}(?:$| .+)'.format(self.name))

        with open(self.dehydrated_domains_path, 'r') as f:
            lines = f.readlines()
        with open(self.dehydrated_domains_path, 'w') as f:
            is_domain_added = False

            for line in lines:
                if not domain_name_re.match(line.strip()):
                    f.write(line)
                elif not remove and not is_domain_added:
                    f.write('%s\n' % self.server_name)
                    is_domain_added = True

            if not remove and not is_domain_added:
                f.write('%s\n' % self.server_name)


# WEB DOMAIN CLASSES
# -----------------------------------------------------------------------------


class NGINXDomain(BaseDomain):
    """
    A Web domain intended to be served by NGINX.

    It is defined by one or more server names, and the protocol on which it
    must be available.
    """

    class Protocol(Enum):
        BOTH = 'both'
        HTTP = 'http'
        HTTPS = 'https'

    nginx_etc_dir = Path('/etc/nginx')
    nginx_log_dir = Path('/var/log/nginx')
    nginx_sites_available_dir = nginx_etc_dir / 'sites-available'
    nginx_sites_enabled_dir = nginx_etc_dir / 'sites-enabled'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._protocol = WebDomain.Protocol.BOTH

    # PROPERTIES

    @property
    def protocol(self) -> Protocol:
        """Return the `Protocol` on which the domain is available."""
        return self._protocol

    @protocol.setter
    def protocol(self, value: Protocol):
        if not isinstance(value, WebDomain.Protocol):
            raise ValueError(
                "Protocol must be a property of `WebDomain.Protocol`."
            )
        self._protocol = value

    # DYNAMIC PROPERTIES

    @cached_property
    def conf_path(self) -> Path:
        """The path to the NGINX configuration of the domain."""
        return self.nginx_sites_available_dir / self.name

    @cached_property
    def enabled_conf_path(self) -> Path:
        """The path to the enabled NGINX configuration of the domain."""
        return self.nginx_sites_enabled_dir / self.name

    @cached_property
    def local_conf_dir(self) -> Path:
        """The directory path to local NGINX configuration of the domain."""
        return self.nginx_sites_available_dir / '{}.d'.format(self.name)

    @cached_property
    def log_dir(self) -> Path:
        """The directory path to NGINX logs of the domain."""
        return self.nginx_log_dir / self.name

    @property
    def has_https(self) -> bool:
        """Whether the domain is available in HTTPS."""
        return self.protocol != WebDomain.Protocol.HTTP

    # OVERRIDABLE METHODS

    def get_conf_context_data(self, **kwargs) -> dict:
        if 'protocol' not in kwargs:
            kwargs['protocol'] = (
                WebDomain.Protocol.HTTP.value
                if not self.has_certs
                else self.protocol.value
            )
        return super().get_conf_context_data(**kwargs)

    def update_properties_from_conf(self, conf: str):
        super().update_properties_from_conf(conf)

        server_name = get_nginx_directive_value('server_name', conf).split(' ')
        if self.name not in server_name:
            raise ConfigurationError(
                "'{}' is not in `server_name {}` directive.".format(
                    self.name, server_name
                ),
                self.conf_path,
            )
        server_name.remove(self.name)
        self.alternative_names = tuple(server_name)

        protocol = self.get_property_from_conf(conf, 'protocol', r'[A-Z]+')
        try:
            self.protocol = WebDomain.Protocol[protocol]
        except KeyError:
            raise ConfigurationError(
                "'{}' is not a valid protocol.".format(protocol),
                self.conf_path,
            )


class WebDomain(DehydratedDomainMixin, NGINXDomain):
    """
    A Web domain intended to be served by NGINX with SSL/TLS certificates.

    It is defined by one or more server names, and the protocol on which it
    must be available. In case of HTTPS, its SSL/TLS certificates can be
    generated using dehydrated.
    """



from pathlib import Path

import pytest

from webdomains.core import (
    BaseDomain,
    ConfigurationError,
    get_nginx_directive_value,
)


class DummyDomain(BaseDomain):
    conf_dir = Path('/etc/dummy/available')
    enabled_conf_dir = Path('/etc/dummy/enabled')

    @property
    def conf_path(self):
        return self.conf_dir / self.name

    @property
    def enabled_conf_path(self):
        return self.enabled_conf_dir / self.name


class TestNGINXHelpers:
    def test_get_nginx_directive_value(self):
        conf = """
server {
  server_name domain.tld www.domain.tld;
}
"""

        assert (
            get_nginx_directive_value('server_name', conf)
            == 'domain.tld www.domain.tld'
        )

    def test_get_nginx_directive_value_error(self):
        with pytest.raises(ConfigurationError) as excinfo:
            get_nginx_directive_value('test', "some configuration")
        assert str(excinfo.value) == "`test` directive is not defined."


class TestBaseDomainConf:
    @pytest.fixture(autouse=True)
    def setup_domain(self):
        self.domain = DummyDomain('test.example.org')

    def test_get_property_from_conf(self):
        conf = """
## protocol=http
## empty=
## spaces = value
 ## indent=value
"""  # noqa: W291

        assert self.domain.get_property_from_conf(conf, 'protocol') == 'http'
        assert self.domain.get_property_from_conf(conf, 'empty') == ''

        with pytest.raises(ConfigurationError):
            self.domain.get_property_from_conf(conf, 'spaces')
        with pytest.raises(ConfigurationError):
            self.domain.get_property_from_conf(conf, 'indent')

    def test_get_property_from_conf_pattern(self):
        conf = """
## protocol=http
## invalid=http1
## empty=
"""

        assert (
            self.domain.get_property_from_conf(conf, 'protocol', r'[a-z]+')
            == 'http'
        )

        with pytest.raises(ConfigurationError):
            self.domain.get_property_from_conf(conf, 'invalid', r'[a-z]+')
        with pytest.raises(ConfigurationError):
            self.domain.get_property_from_conf(conf, 'empty', r'.+')

    def test_get_property_from_conf_not_defined(self):
        with pytest.raises(ConfigurationError) as excinfo:
            self.domain.get_property_from_conf("some configuration", 'test')
        assert excinfo.value.message == "`test` variable is not defined."



import logging
import os
import threading

import click


class cached_property:
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    A cached property can be made out of an existing method:

        url = cached_property(get_absolute_url)

    Based on `django.utils.functional.cached_property`.
    """

    name = None

    @staticmethod
    def func(instance):
        raise TypeError(
            "Cannot use cached_property instance without calling "
            "__set_name__() on it."
        )

    def __init__(self, func):
        self.real_func = func
        self.__doc__ = getattr(func, '__doc__')

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name
            self.func = self.real_func
        elif name != self.name:
            raise TypeError(
                "Cannot assign the same cached_property to two different "
                "names (%r and %r)." % (self.name, name)
            )

    def __get__(self, instance, cls=None):
        """
        Call the function and put the return value in instance.__dict__ so that
        subsequent attribute access on the instance returns the cached value
        instead of calling cached_property.__get__().
        """
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


# LOGGING
# -----------------------------------------------------------------------------


class LogPipe(threading.Thread):
    """
    A pipe which outputs data to a logger.

    It is intended to be used with `subprocess` as the `stdout` or `stderr`
    arguments, i.e.:

        with LogPipe(logging.DEBUG) as stdout:
            subprocess.run(['ls', '-l', '/'], stdout=stdout)
    """

    def __init__(self, level, logger=None):
        super().__init__(daemon=False)

        self.level = level
        self.logger = logger or logging
        self.pread, self.pwrite = os.pipe()
        self.preader = os.fdopen(self.pread)

        self.start()

    def fileno(self):
        """Return the write file descriptor of the pipe."""
        return self.pwrite

    def close(self):
        """Close the write file descriptor of the pipe."""
        os.close(self.pwrite)

    def run(self):
        for line in iter(self.preader.readline, ''):
            self.logger.log(self.level, line.strip('\n'))
        self.preader.close()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()


class ClickHandler(logging.Handler):
    """Handler which sends logging output to `click.echo`."""

    def emit(self, record):
        try:
            click.echo(
                self.format(record), err=record.levelno >= logging.ERROR
            )
        except Exception:
            self.handleError(record)


class ColorFormatter(logging.Formatter):
    """Formatter which extends `LogRecord` attributes with 'levelprefix'.

    The 'levelprefix' attribute is almost a colored version of 'levelname', see
    `ColorFormatter.STYLES` for the styles attributes.
    """

    STYLES = {
        'ERROR': dict(text='Error! ', fg='red'),
        'EXCEPTION': dict(text='Error! ', fg='red'),
        'CRITICAL': dict(text='Error! ', fg='red'),
        'INFO': dict(text=' + ', fg='blue'),
        'WARNING': dict(text='Warning: ', fg='yellow'),
    }

    def formatMessage(self, record):
        style_kwargs = self.STYLES.get(record.levelname, {})
        style_kwargs.setdefault('text', '')
        record.levelprefix = click.style(**style_kwargs)
        return super().formatMessage(record)


def setup_click_logger(logger, fmt=None):
    """Setup a logger to be used with `click`."""
    formatter = ColorFormatter(fmt or '%(levelprefix)s%(message)s')
    handler = ClickHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger
